@@ -10,16 +10,19 @@ AST node type to code that actually does the bulk of the work. For
 example, expressions are transformed in mypyc.irbuild.expression and
 functions are transformed in mypyc.irbuild.function.
 """
+from contextlib import contextmanager
 
-from typing import Callable, Dict, List, Tuple, Optional, Union, Sequence, Set, Any
-from typing_extensions import overload
-from mypy.ordered_dict import OrderedDict
+from mypyc.irbuild.prepare import RegisterImplInfo
+from typing import Callable, Dict, List, Tuple, Optional, Union, Sequence, Set, Any, Iterator
+from typing_extensions import overload, Final
+from mypy.backports import OrderedDict
 
 from mypy.build import Graph
 from mypy.nodes import (
     MypyFile, SymbolNode, Statement, OpExpr, IntExpr, NameExpr, LDEF, Var, UnaryExpr,
     CallExpr, IndexExpr, Expression, MemberExpr, RefExpr, Lvalue, TupleExpr,
-    TypeInfo, Decorator, OverloadedFuncDef, StarExpr, ComparisonExpr, GDEF, ARG_POS, ARG_NAMED
+    TypeInfo, Decorator, OverloadedFuncDef, StarExpr,
+    GDEF, ArgKind, ARG_POS, ARG_NAMED, FuncDef,
 )
 from mypy.types import (
     Type, Instance, TupleType, UninhabitedType, get_proper_type
@@ -35,9 +38,9 @@ from mypyc.ir.ops import (
     SetAttr, LoadStatic, InitStatic, NAMESPACE_MODULE, RaiseStandardError
 )
 from mypyc.ir.rtypes import (
-    RType, RTuple, RInstance, int_rprimitive, dict_rprimitive,
+    RType, RTuple, RInstance, c_int_rprimitive, int_rprimitive, dict_rprimitive,
     none_rprimitive, is_none_rprimitive, object_rprimitive, is_object_rprimitive,
-    str_rprimitive, is_tagged, is_list_rprimitive, is_tuple_rprimitive, c_pyssize_t_rprimitive
+    str_rprimitive, is_list_rprimitive, is_tuple_rprimitive, c_pyssize_t_rprimitive
 )
 from mypyc.ir.func_ir import FuncIR, INVALID_FUNC_DEF, RuntimeArg, FuncSignature, FuncDecl
 from mypyc.ir.class_ir import ClassIR, NonExtClassInfo
@@ -45,7 +48,9 @@ from mypyc.primitives.registry import CFunctionDescription, function_ops
 from mypyc.primitives.list_ops import to_list, list_pop_last, list_get_item_unsafe_op
 from mypyc.primitives.dict_ops import dict_get_item_op, dict_set_item_op
 from mypyc.primitives.generic_ops import py_setattr_op, iter_op, next_op
-from mypyc.primitives.misc_ops import import_op, check_unpack_count_op, get_module_dict_op
+from mypyc.primitives.misc_ops import (
+    import_op, check_unpack_count_op, get_module_dict_op, import_extra_args_op
+)
 from mypyc.crash import catch_errors
 from mypyc.options import CompilerOptions
 from mypyc.errors import Errors
@@ -62,6 +67,11 @@ from mypyc.irbuild.ll_builder import LowLevelIRBuilder
 from mypyc.irbuild.util import is_constant
 
 
+# These int binary operations can borrow their operands safely, since the
+# primitives take this into consideration.
+int_borrow_friendly_op: Final = {'+', '-', '==', '!=', '<', '<=', '>', '>='}
+
+
 class IRVisitor(ExpressionVisitor[Value], StatementVisitor[None]):
     pass
 
@@ -82,23 +92,24 @@ class IRBuilder:
                  mapper: Mapper,
                  pbv: PreBuildVisitor,
                  visitor: IRVisitor,
-                 options: CompilerOptions) -> None:
+                 options: CompilerOptions,
+                 singledispatch_impls: Dict[FuncDef, List[RegisterImplInfo]]) -> None:
         self.builder = LowLevelIRBuilder(current_module, mapper, options)
         self.builders = [self.builder]
-        self.symtables = [OrderedDict()]  # type: List[OrderedDict[SymbolNode, SymbolTarget]]
-        self.runtime_args = [[]]  # type: List[List[RuntimeArg]]
-        self.function_name_stack = []  # type: List[str]
-        self.class_ir_stack = []  # type: List[ClassIR]
+        self.symtables: List[OrderedDict[SymbolNode, SymbolTarget]] = [OrderedDict()]
+        self.runtime_args: List[List[RuntimeArg]] = [[]]
+        self.function_name_stack: List[str] = []
+        self.class_ir_stack: List[ClassIR] = []
 
         self.current_module = current_module
         self.mapper = mapper
         self.types = types
         self.graph = graph
-        self.ret_types = []  # type: List[RType]
-        self.functions = []  # type: List[FuncIR]
-        self.classes = []  # type: List[ClassIR]
-        self.final_names = []  # type: List[Tuple[str, RType]]
-        self.callable_class_names = set()  # type: Set[str]
+        self.ret_types: List[RType] = []
+        self.functions: List[FuncIR] = []
+        self.classes: List[ClassIR] = []
+        self.final_names: List[Tuple[str, RType]] = []
+        self.callable_class_names: Set[str] = set()
         self.options = options
 
         # These variables keep track of the number of lambdas, implicit indices, and implicit
@@ -113,6 +124,7 @@ class IRBuilder:
         self.encapsulating_funcs = pbv.encapsulating_funcs
         self.nested_fitems = pbv.nested_funcs.keys()
         self.fdefs_to_decorators = pbv.funcs_to_decorators
+        self.singledispatch_impls = singledispatch_impls
 
         self.visitor = visitor
 
@@ -122,17 +134,19 @@ class IRBuilder:
         # be generated) is stored in that FuncInfo instance. When the function is done being
         # generated, its corresponding FuncInfo is popped off the stack.
         self.fn_info = FuncInfo(INVALID_FUNC_DEF, '', '')
-        self.fn_infos = [self.fn_info]  # type: List[FuncInfo]
+        self.fn_infos: List[FuncInfo] = [self.fn_info]
 
         # This list operates as a stack of constructs that modify the
         # behavior of nonlocal control flow constructs.
-        self.nonlocal_control = []  # type: List[NonlocalControl]
+        self.nonlocal_control: List[NonlocalControl] = []
 
         self.errors = errors
         # Notionally a list of all of the modules imported by the
         # module being compiled, but stored as an OrderedDict so we
         # can also do quick lookups.
-        self.imports = OrderedDict()  # type: OrderedDict[str, None]
+        self.imports: OrderedDict[str, None] = OrderedDict()
+
+        self.can_borrow = False
 
     # High-level control
 
@@ -145,15 +159,23 @@ class IRBuilder:
         self.module_path = module_path
 
     @overload
-    def accept(self, node: Expression) -> Value: ...
+    def accept(self, node: Expression, *, can_borrow: bool = False) -> Value: ...
 
     @overload
     def accept(self, node: Statement) -> None: ...
 
-    def accept(self, node: Union[Statement, Expression]) -> Optional[Value]:
-        """Transform an expression or a statement."""
+    def accept(self, node: Union[Statement, Expression], *,
+               can_borrow: bool = False) -> Optional[Value]:
+        """Transform an expression or a statement.
+
+        If can_borrow is true, prefer to generate a borrowed reference.
+        Borrowed references are faster since they don't require reference count
+        manipulation, but they are only safe to use in specific contexts.
+        """
         with self.catch_errors(node.line):
             if isinstance(node, Expression):
+                old_can_borrow = self.can_borrow
+                self.can_borrow = can_borrow
                 try:
                     res = node.accept(self.visitor)
                     res = self.coerce(res, self.node_type(node), node.line)
@@ -163,6 +185,9 @@ class IRBuilder:
                 # from causing more downstream trouble.
                 except UnsupportedException:
                     res = Register(self.node_type(node))
+                self.can_borrow = old_can_borrow
+                if not can_borrow:
+                    self.flush_keep_alives()
                 return res
             else:
                 try:
@@ -171,6 +196,9 @@ class IRBuilder:
                     pass
                 return None
 
+    def flush_keep_alives(self) -> None:
+        self.builder.flush_keep_alives()
+
     # Pass through methods for the most common low-level builder ops, for convenience.
 
     def add(self, op: Op) -> Value:
@@ -194,6 +222,16 @@ class IRBuilder:
     def load_str(self, value: str) -> Value:
         return self.builder.load_str(value)
 
+    def load_bytes_from_str_literal(self, value: str) -> Value:
+        """Load bytes object from a string literal.
+
+        The literal characters of BytesExpr (the characters inside b'')
+        are stored in BytesExpr.value, whose type is 'str' not 'bytes'.
+        Thus we perform a special conversion here.
+        """
+        bytes_value = bytes(value, 'utf8').decode('unicode-escape').encode('raw-unicode-escape')
+        return self.builder.load_bytes(bytes_value)
+
     def load_int(self, value: int) -> Value:
         return self.builder.load_int(value)
 
@@ -204,7 +242,7 @@ class IRBuilder:
         return self.builder.binary_op(lreg, rreg, expr_op, line)
 
     def coerce(self, src: Value, target_type: RType, line: int, force: bool = False) -> Value:
-        return self.builder.coerce(src, target_type, line, force)
+        return self.builder.coerce(src, target_type, line, force, can_borrow=self.can_borrow)
 
     def none_object(self) -> Value:
         return self.builder.none_object()
@@ -235,7 +273,7 @@ class IRBuilder:
                 function: Value,
                 arg_values: List[Value],
                 line: int,
-                arg_kinds: Optional[List[int]] = None,
+                arg_kinds: Optional[List[ArgKind]] = None,
                 arg_names: Optional[Sequence[Optional[str]]] = None) -> Value:
         return self.builder.py_call(function, arg_values, line, arg_kinds, arg_names)
 
@@ -251,10 +289,10 @@ class IRBuilder:
                         arg_values: List[Value],
                         result_type: Optional[RType],
                         line: int,
-                        arg_kinds: Optional[List[int]] = None,
+                        arg_kinds: Optional[List[ArgKind]] = None,
                         arg_names: Optional[List[Optional[str]]] = None) -> Value:
         return self.builder.gen_method_call(
-            base, name, arg_values, result_type, line, arg_kinds, arg_names
+            base, name, arg_values, result_type, line, arg_kinds, arg_names, self.can_borrow
         )
 
     def load_module(self, name: str) -> Value:
@@ -286,19 +324,45 @@ class IRBuilder:
         key_unicode = self.load_str(key)
         self.call_c(dict_set_item_op, [non_ext.dict, key_unicode, val], line)
 
+    def gen_import_from(self, id: str, globals_dict: Value,
+                        imported: List[str], line: int) -> Value:
+        self.imports[id] = None
+
+        null_dict = Integer(0, dict_rprimitive, line)
+        names_to_import = self.new_list_op([self.load_str(name) for name in imported], line)
+        zero_int = Integer(0, c_int_rprimitive, line)
+        value = self.call_c(
+            import_extra_args_op,
+            [self.load_str(id), globals_dict, null_dict, names_to_import, zero_int],
+            line,
+        )
+        self.add(InitStatic(value, id, namespace=NAMESPACE_MODULE))
+        return value
+
     def gen_import(self, id: str, line: int) -> None:
         self.imports[id] = None
 
         needs_import, out = BasicBlock(), BasicBlock()
-        first_load = self.load_module(id)
-        comparison = self.translate_is_op(first_load, self.none_object(), 'is not', line)
-        self.add_bool_branch(comparison, out, needs_import)
+        self.check_if_module_loaded(id, line, needs_import, out)
 
         self.activate_block(needs_import)
         value = self.call_c(import_op, [self.load_str(id)], line)
         self.add(InitStatic(value, id, namespace=NAMESPACE_MODULE))
         self.goto_and_activate(out)
 
+    def check_if_module_loaded(self, id: str, line: int,
+                               needs_import: BasicBlock, out: BasicBlock) -> None:
+        """Generate code that checks if the module `id` has been loaded yet.
+
+        Arguments:
+            id: name of module to check if imported
+            line: line number that the import occurs on
+            needs_import: the BasicBlock that is run if the module has not been loaded yet
+            out: the BasicBlock that is run if the module has already been loaded"""
+        first_load = self.load_module(id)
+        comparison = self.translate_is_op(first_load, self.none_object(), 'is not', line)
+        self.add_bool_branch(comparison, out, needs_import)
+
     def get_module(self, module: str, line: int) -> Value:
         # Python 3.7 has a nice 'PyImport_GetModule' function that we can't use :(
         mod_dict = self.call_c(get_module_dict_op, [], line)
@@ -376,7 +440,7 @@ class IRBuilder:
             if class_name is None:
                 name = lvalue.name
             else:
-                name = '{}.{}'.format(class_name, lvalue.name)
+                name = f'{class_name}.{lvalue.name}'
             assert name is not None, "Full name not set for variable"
             coerced = self.coerce(rvalue_reg, type_override or self.node_type(lvalue), lvalue.line)
             self.final_names.append((name, coerced.type))
@@ -389,7 +453,7 @@ class IRBuilder:
         module, name = split_name
         return self.builder.load_static_checked(
             typ, name, module, line=line,
-            error_msg='value for final name "{}" was not set'.format(error_name))
+            error_msg=f'value for final name "{error_name}" was not set')
 
     def load_final_literal_value(self, val: Union[int, str, bytes, float, bool],
                                  line: int) -> Value:
@@ -454,11 +518,12 @@ class IRBuilder:
             return AssignmentTargetIndex(base, index)
         elif isinstance(lvalue, MemberExpr):
             # Attribute assignment x.y = e
-            obj = self.accept(lvalue.expr)
-            return AssignmentTargetAttr(obj, lvalue.name)
+            can_borrow = self.is_native_attr_ref(lvalue)
+            obj = self.accept(lvalue.expr, can_borrow=can_borrow)
+            return AssignmentTargetAttr(obj, lvalue.name, can_borrow=can_borrow)
         elif isinstance(lvalue, TupleExpr):
             # Multiple assignment a, ..., b = e
-            star_idx = None  # type: Optional[int]
+            star_idx: Optional[int] = None
             lvalues = []
             for idx, item in enumerate(lvalue.items):
                 targ = self.get_assignment_target(item)
@@ -475,7 +540,10 @@ class IRBuilder:
 
         assert False, 'Unsupported lvalue: %r' % lvalue
 
-    def read(self, target: Union[Value, AssignmentTarget], line: int = -1) -> Value:
+    def read(self,
+             target: Union[Value, AssignmentTarget],
+             line: int = -1,
+             can_borrow: bool = False) -> Value:
         if isinstance(target, Value):
             return target
         if isinstance(target, AssignmentTargetRegister):
@@ -488,7 +556,8 @@ class IRBuilder:
             assert False, target.base.type
         if isinstance(target, AssignmentTargetAttr):
             if isinstance(target.obj.type, RInstance) and target.obj.type.class_ir.is_ext_class:
-                return self.add(GetAttr(target.obj, target.attr, line))
+                borrow = can_borrow and target.can_borrow
+                return self.add(GetAttr(target.obj, target.attr, line, borrow=borrow))
             else:
                 return self.py_get_attr(target.obj, target.attr, line)
 
@@ -499,7 +568,7 @@ class IRBuilder:
                rvalue_reg: Value,
                line: int) -> None:
         if isinstance(target, Register):
-            self.add(Assign(target, rvalue_reg))
+            self.add(Assign(target, self.coerce(rvalue_reg, target.type, line)))
         elif isinstance(target, AssignmentTargetRegister):
             rvalue_reg = self.coerce(rvalue_reg, target.type, line)
             self.add(Assign(target.register, rvalue_reg))
@@ -642,7 +711,7 @@ class IRBuilder:
 
     def spill(self, value: Value) -> AssignmentTarget:
         """Moves a given Value instance into the generator class' environment class."""
-        name = '{}{}'.format(TEMP_ATTR_NAME, self.temp_counter)
+        name = f'{TEMP_ATTR_NAME}{self.temp_counter}'
         self.temp_counter += 1
         target = self.add_var_to_env_class(Var(name), value.type, self.fn_info.generator_class)
         # Shouldn't be able to fail, so -1 for line
@@ -729,7 +798,7 @@ class IRBuilder:
 
         from mypy.join import join_types
         if isinstance(iterable, TupleType):
-            joined = UninhabitedType()  # type: Type
+            joined: Type = UninhabitedType()
             for item in iterable.items:
                 joined = join_types(joined, item)
             return joined
@@ -774,7 +843,7 @@ class IRBuilder:
                 is_final = sym.node.is_final or expr_fullname == 'enum.Enum'
                 if is_final:
                     final_var = sym.node
-                    fullname = '{}.{}'.format(sym.node.info.fullname, final_var.name)
+                    fullname = f'{sym.node.info.fullname}.{final_var.name}'
                     native = self.is_native_module(expr.expr.node.module_name)
         elif self.is_module_member_expr(expr):
             # a module attribute
@@ -825,6 +894,16 @@ class IRBuilder:
         callee_node = callee.node
         if isinstance(callee_node, OverloadedFuncDef):
             callee_node = callee_node.impl
+        # TODO: use native calls for any decorated functions which have all their decorators
+        # removed, not just singledispatch functions (which we don't do now just in case those
+        # decorated functions are callable classes or cannot be called without the python API for
+        # some other reason)
+        if (
+            isinstance(callee_node, Decorator)
+            and callee_node.func not in self.fdefs_to_decorators
+            and callee_node.func in self.singledispatch_impls
+        ):
+            callee_node = callee_node.func
         if (callee_node is not None
                 and callee.fullname is not None
                 and callee_node in self.mapper.func_to_decl
@@ -845,61 +924,6 @@ class IRBuilder:
             expr.line
         )
 
-    # Conditional expressions
-
-    def process_conditional(self, e: Expression, true: BasicBlock, false: BasicBlock) -> None:
-        if isinstance(e, OpExpr) and e.op in ['and', 'or']:
-            if e.op == 'and':
-                # Short circuit 'and' in a conditional context.
-                new = BasicBlock()
-                self.process_conditional(e.left, new, false)
-                self.activate_block(new)
-                self.process_conditional(e.right, true, false)
-            else:
-                # Short circuit 'or' in a conditional context.
-                new = BasicBlock()
-                self.process_conditional(e.left, true, new)
-                self.activate_block(new)
-                self.process_conditional(e.right, true, false)
-        elif isinstance(e, UnaryExpr) and e.op == 'not':
-            self.process_conditional(e.expr, false, true)
-        else:
-            res = self.maybe_process_conditional_comparison(e, true, false)
-            if res:
-                return
-            # Catch-all for arbitrary expressions.
-            reg = self.accept(e)
-            self.add_bool_branch(reg, true, false)
-
-    def maybe_process_conditional_comparison(self,
-                                             e: Expression,
-                                             true: BasicBlock,
-                                             false: BasicBlock) -> bool:
-        """Transform simple tagged integer comparisons in a conditional context.
-
-        Return True if the operation is supported (and was transformed). Otherwise,
-        do nothing and return False.
-
-        Args:
-            e: Arbitrary expression
-            true: Branch target if comparison is true
-            false: Branch target if comparison is false
-        """
-        if not isinstance(e, ComparisonExpr) or len(e.operands) != 2:
-            return False
-        ltype = self.node_type(e.operands[0])
-        rtype = self.node_type(e.operands[1])
-        if not is_tagged(ltype) or not is_tagged(rtype):
-            return False
-        op = e.operators[0]
-        if op not in ('==', '!=', '<', '<=', '>', '>='):
-            return False
-        left = self.accept(e.operands[0])
-        right = self.accept(e.operands[1])
-        # "left op right" for two tagged integers
-        self.builder.compare_tagged_condition(left, right, op, true, false, e.line)
-        return True
-
     # Basic helpers
 
     def flatten_classes(self, arg: Union[RefExpr, TupleExpr]) -> Optional[List[ClassIR]]:
@@ -914,7 +938,7 @@ class IRBuilder:
                     return [ir]
             return None
         else:
-            res = []  # type: List[ClassIR]
+            res: List[ClassIR] = []
             for item in arg.items:
                 if isinstance(item, (RefExpr, TupleExpr)):
                     item_part = self.flatten_classes(item)
@@ -952,22 +976,18 @@ class IRBuilder:
         self.fn_info = self.fn_infos[-1]
         return builder.args, runtime_args, builder.blocks, ret_type, fn_info
 
+    @contextmanager
     def enter_method(self,
                      class_ir: ClassIR,
                      name: str,
                      ret_type: RType,
                      fn_info: Union[FuncInfo, str] = '',
-                     self_type: Optional[RType] = None) -> None:
-        """Begin generating IR for a method.
+                     self_type: Optional[RType] = None) -> Iterator[None]:
+        """Generate IR for a method.
 
         If the method takes arguments, you should immediately afterwards call
         add_argument() for each non-self argument (self is created implicitly).
 
-        Call leave_method() to finish the generation of the method.
-
-        You can enter multiple methods at a time. They are maintained in a
-        stack, and leave_method() leaves the topmost one.
-
         Args:
             class_ir: Add method to this class
             name: Short name of the method
@@ -983,8 +1003,20 @@ class IRBuilder:
         if self_type is None:
             self_type = RInstance(class_ir)
         self.add_argument(SELF_NAME, self_type)
+        try:
+            yield
+        finally:
+            arg_regs, args, blocks, ret_type, fn_info = self.leave()
+            sig = FuncSignature(args, ret_type)
+            name = self.function_name_stack.pop()
+            class_ir = self.class_ir_stack.pop()
+            decl = FuncDecl(name, class_ir.name, self.module_name, sig)
+            ir = FuncIR(decl, arg_regs, blocks)
+            class_ir.methods[name] = ir
+            class_ir.method_decls[name] = ir.decl
+            self.functions.append(ir)
 
-    def add_argument(self, var: Union[str, Var], typ: RType, kind: int = ARG_POS) -> Register:
+    def add_argument(self, var: Union[str, Var], typ: RType, kind: ArgKind = ARG_POS) -> Register:
         """Declare an argument in the current function.
 
         You should use this instead of directly calling add_local() in new code.
@@ -995,18 +1027,6 @@ class IRBuilder:
         self.runtime_args[-1].append(RuntimeArg(var.name, typ, kind))
         return reg
 
-    def leave_method(self) -> None:
-        """Finish the generation of IR for a method."""
-        arg_regs, args, blocks, ret_type, fn_info = self.leave()
-        sig = FuncSignature(args, ret_type)
-        name = self.function_name_stack.pop()
-        class_ir = self.class_ir_stack.pop()
-        decl = FuncDecl(name, class_ir.name, self.module_name, sig)
-        ir = FuncIR(decl, arg_regs, blocks)
-        class_ir.methods[name] = ir
-        class_ir.method_decls[name] = ir.decl
-        self.functions.append(ir)
-
     def lookup(self, symbol: SymbolNode) -> SymbolTarget:
         return self.symtables[-1][symbol]
 
@@ -1017,7 +1037,12 @@ class IRBuilder:
             is_arg: is this a function argument
         """
         assert isinstance(symbol, SymbolNode)
-        reg = Register(typ, symbol.name, is_arg=is_arg, line=symbol.line)
+        reg = Register(
+            typ,
+            remangle_redefinition_name(symbol.name),
+            is_arg=is_arg,
+            line=symbol.line,
+        )
         self.symtables[-1][symbol] = AssignmentTargetRegister(reg)
         if is_arg:
             self.builder.args.append(reg)
@@ -1109,6 +1134,14 @@ class IRBuilder:
         left = self.load_module(module)
         return self.py_get_attr(left, name, line)
 
+    def is_native_attr_ref(self, expr: MemberExpr) -> bool:
+        """Is expr a direct reference to a native (struct) attribute of an instance?"""
+        obj_rtype = self.node_type(expr.expr)
+        return (isinstance(obj_rtype, RInstance)
+                and obj_rtype.class_ir.is_ext_class
+                and obj_rtype.class_ir.has_attr(expr.name)
+                and not obj_rtype.class_ir.get_method(expr.name))
+
     # Lacks a good type because there wasn't a reasonable type in 3.5 :(
     def catch_errors(self, line: int) -> Any:
         return catch_errors(self.module_path, line)
@@ -1119,6 +1152,9 @@ class IRBuilder:
     def error(self, msg: str, line: int) -> None:
         self.errors.error(msg, self.module_path, line)
 
+    def note(self, msg: str, line: int) -> None:
+        self.errors.note(msg, self.module_path, line)
+
 
 def gen_arg_defaults(builder: IRBuilder) -> None:
     """Generate blocks for arguments that have default values.
@@ -1151,3 +1187,14 @@ def gen_arg_defaults(builder: IRBuilder) -> None:
                         GetAttr(builder.fn_info.callable_class.self_reg, name, arg.line))
             assert isinstance(target, AssignmentTargetRegister)
             builder.assign_if_null(target.register, get_default, arg.initializer.line)
+
+
+def remangle_redefinition_name(name: str) -> str:
+    """Remangle names produced by mypy when allow-redefinition is used and a name
+    is used with multiple types within a single block.
+
+    We only need to do this for locals, because the name is used as the name of the register;
+    for globals, the name itself is stored in a register for the purpose of doing dict
+    lookups.
+    """
+    return name.replace("'", "__redef__")
